# Run this file, using a marshalled version of the RPC protocol on file descriptor #3.
# Kill it when done.
from __future__ import absolute_import
import os, sys, marshal, threading, json, importlib
from anvil_downlink_worker import handle_incoming_call, load_app
from anvil import _serialise, _server, _threaded_server
import anvil.server
import anvil.pdf

PIPE_IN = os.fdopen(os.dup(0), 'rb')
PIPE_OUT = os.fdopen(os.dup(1), 'wb')
OLD_STDERR = os.fdopen(os.dup(2), 'wb')
WRITE_LOCK = threading.RLock()

_threaded_server.console_output = OLD_STDERR

# We use stdin and stdout to talk Anvil-RPC to our manager process.
# We overwrite sys.stdout and sys.stderr with shims that produce 'output' updates across
# that interface, and to make sure no direct use of FDs 0/1/2 get in its way
# To make sure no clever use Overwrite actual stdin and stdout with a thread that forwards them
new_stdout = os.open(os.devnull, os.O_RDWR)
os.dup2(new_stdout, 0)
os.dup2(new_stdout, 1)
#os.dup2(new_stdout, 2)

def write_pipe(data, bin=None):
    with WRITE_LOCK:
        try:
            s = b'X' + marshal.dumps(data)
            if bin is not None:
                s += marshal.dumps(bin)
        except ValueError:
            raise _server.SerializationError("You can only pass strings, numbers, arrays, lists, LiveObjects and Media to or from server functions")

        PIPE_OUT.write(s)
        PIPE_OUT.flush()


class DummyStdout:
    def write(self, s):
        write_pipe({'output': s, 'id': _threaded_server.call_info.call_id})
    def flush(self):
        pass


sys.stdout = DummyStdout()
#sys.stderr = sys.stdout

def dbg(s):
    sys.stderr.write(s+"\n")
    sys.stderr.flush()


def send_reqresp(r, collect_capabilities=None):
    _serialise.serialise(r, write_pipe, collect_capabilities=collect_capabilities)

_threaded_server.send_reqresp = send_reqresp


def run():
    while True:
        PIPE_IN.read(1) # signal byte to wake us up
        msg = marshal.load(PIPE_IN)
        type = msg.get("type", None)
        if type in ["CALL_WITH_APP", "LAUNCH_BACKGROUND_WITH_APP", "CALL", "LAUNCH_BACKGROUND",
                    "LAUNCH_REPL", "REPL_COMMAND", "TERMINATE_REPL"]:
            handle_incoming_call(msg, write_pipe)

        elif type == "PROVIDE_APP":
            load_app(msg["app"])

        elif type == "GET_TASK_STATE":
            def err(*args):
                raise _server.SerializationError("Cannot use BlobMedia objects in task state.")

            try:
                sjson = _server.fill_out_media({'id': msg['id'], 'response': anvil.server.task_state}, err)
                json.dumps(sjson)
            except (TypeError, _server.SerializationError) as e:
                write_pipe({'id': msg['id'], 'error': {'type': 'anvil.server.SerializationError', 'message': "Illegal value in a anvil.server.task_state. " + e.args[0]}})
            except Exception as e:
                write_pipe({'id': msg['id'], 'error': {'type': 'anvil.server.InternalError', 'message': "Could not get task state: " + e.args[0]}})
            else:
                write_pipe(sjson)

        elif type == "CHUNK_HEADER":
            _serialise.process_blob_header(msg)
            _serialise.process_blob(marshal.load(PIPE_IN))
        elif type is None and ("response" in msg or "error" in msg):
            _threaded_server.IncomingResponse(msg)
        else:
            print("Downlink worker socket got unrecognised message: "+repr(msg))
            sys.stdout.flush()
            os._exit(1)


def plotly_serialization_helper(class_fullname):
    name_parts = class_fullname.split(".")
    module_name = ".".join(name_parts[:-1])
    class_name = name_parts[-1]

    module = importlib.import_module(module_name)
    cls = getattr(module, class_name)

    if not hasattr(cls, '__serialize__'):
        #print(f"Registering {cls}")
        def serialize(self, global_data):
            #print("Serialising %s on downlink" % type(self))
            return self.to_plotly_json()

        @staticmethod
        def new_deserialized(data, global_data):
            #print("Deserialising %s on downlink" % cls)
            return cls(data)

        cls.__serialize__ = serialize
        cls.__new_deserialized__ = new_deserialized
        anvil.server.portable_class(cls, class_fullname)

_server._serialization_helpers["plotly.graph_objs"] = plotly_serialization_helper


if __name__ == "__main__":
    run()



import importlib
from . import _components
from . import _server

# Filled in by the downlink_worker package
packages_by_app_id = {}


def mk_component(yaml, components_by_name):
    component_type = yaml['type']
    if component_type.startswith('form:'):

        spec = component_type.split(":")
        if len(spec) == 3:
            app_id = spec[1]
            form_name = spec[2]
        elif len(spec) == 2:
            app_id = ''
            form_name = spec[1]
        else:
            raise ValueError("Can't instantiate custom component %s" % component_type)

        form_mod = importlib.import_module(packages_by_app_id[app_id]+"."+form_name)
        form_cls = getattr(form_mod, form_name.split(".")[-1])

        return form_cls(**yaml.get('properties', {}))
    else:
        try:
            cls = getattr(_components, component_type)
        except AttributeError:
            # Okay, make a dummy
            obj = _components.Container() if 'components' in yaml else _components.Component()
            obj.SERIALIZATION_INFO = "anvil." + component_type, type(obj)
        else:
            obj = cls(**yaml['properties'])

        if isinstance(obj, _components.Container):
            for c in yaml.get('components', []):
                obj.add_component(mk_component(c, components_by_name), **c.get('layout_properties', {}))

        components_by_name[yaml['name']] = obj

        return obj


def init_components_on_form(form, components):
    components_by_name = {}
    for c in components:
        form.add_component(mk_component(c, components_by_name), **c.get('layout_properties', {}))

    for n,c in components_by_name.items():
        setattr(form, n, c)


def mk_template_class(form_yaml):
    ns = {}
    container_cls = getattr(_components, form_yaml['container']['type'])
    default_custom_props = {}
    if form_yaml.get('custom_component'):
        for pt in form_yaml.get('properties', []):
            if 'default_value' in pt:
                default_custom_props[pt['name']] = pt['default_value']

    def __new__(cls, **kwargs):
        container_new = container_cls.__new__

        # This is a hack using the fact that component properties don't actually have any behaviour
        # on the server
        obj = container_new(cls)
        obj.__dict__["$_components"] = []
        obj.__dict__["$container_props"] = dict(form_yaml['container']['properties'])

        # No event binding
        # No data bindings

        container_cls.__init__(obj, __ignore_property_exceptions=True, **kwargs)

        init_components_on_form(obj, form_yaml['components'])

        return obj

    ns['__new__'] = __new__

    def __getattr__(self, name):
        if name in default_custom_props:
            return default_custom_props[name]

        cp = self.__dict__["$container_props"]
        try:
            return cp[name]
        except KeyError:
            raise AttributeError(name)

    ns['__getattr__'] = __getattr__

    def __setattr__(self, name, value):
        cp = self.__dict__["$container_props"]
        if name in cp:
            cp[name] = value
        else:
            object.__setattr__(self, name, value)

    ns['__setattr__'] = __setattr__

    def __init__(self, __ignore_property_exceptions=False, **kwargs):
        for n,v in kwargs.items():
            setattr(self, n, v)

    ns['__init__'] = ns['init_components'] = __init__

    def __serialize_once__(self, global_data):
        d = dict(self.__dict__)
        components = d.pop("$_components")
        container_props = d.pop("$container_props")

        return {"d": d, "a": container_props, "c": components}

    ns['__serialize_once__'] = __serialize_once__

    def __deserialize__(self, data, global_data):
        self.__dict__['$_components'] = data['c']
        self.__dict__['$container_props'] = data['a']
        self.__dict__.update(data['d'])

    ns['__deserialize__'] = __deserialize__

    ns['refresh_data_bindings'] = lambda: None

    template_name = str(form_yaml['class_name'].split(".")[-1]) + "Template"

    return type(template_name, (container_cls, _server.SerializeWithIdentity), ns)



from anvil.server import portable_class

# Hack: Force ourselves into the top-level package, even
# if we were loaded into a runtime-v1 per-app Anvil package
__package__ = "anvil.tables"
__name__ = "anvil.tables.query"

#!defFunction(anvil.tables.query,_,pattern)!2: "Match values using a case-sensitive LIKE query, using the % wildcard character." ["like"]
@portable_class
class like(object):
    def __init__(self, pattern):
        self.pattern = pattern

#!defFunction(anvil.tables.query,_,pattern)!2: "Match values using a case-insensitive ILIKE query, using the % wildcard character." ["ilike"]
@portable_class
class ilike(object):
    def __init__(self, pattern):
        self.pattern = pattern

#!defFunction(anvil.tables.query,_,value)!2: "Match values greater than the provided value." ["greater_than"]
@portable_class
class greater_than(object):
    def __init__(self, value):
        self.value = value

#!defFunction(anvil.tables.query,_,value)!2: "Match values less than the provided value." ["less_than"]
@portable_class
class less_than(object):
    def __init__(self, value):
        self.value = value

#!defFunction(anvil.tables.query,_,value)!2: "Match values greater than or equal to the provided value." ["greater_than_or_equal_to"]
@portable_class
class greater_than_or_equal_to(object):
    def __init__(self, value):
        self.value = value

#!defFunction(anvil.tables.query,_,value)!2: "Match values less than or equal to the provided value." ["less_than_or_equal_to"]
@portable_class
class less_than_or_equal_to(object):
    def __init__(self, value):
        self.value = value

#!defFunction(anvil.tables.query,_,min,max,[min_inclusive=True],[max_inclusive=False])!2: "Match values between the provided min and max, optionally inclusive." ["between"]
def between(min, max, min_inclusive=True, max_inclusive=False):
    return all_of(
        greater_than_or_equal_to(min) if min_inclusive else greater_than(min),
        less_than_or_equal_to(max) if max_inclusive else less_than(max)
    )

#!defFunction(anvil.tables.query,_,query,[raw=False])!2: "Match values that match the provided full-text search query." ["full_text_match"]
@portable_class
class full_text_match(object):
    def __init__(self, query, raw=False):
        self.query = query
        self.raw = raw

#!defFunction(anvil.tables.query,_,*query_expressions)!2: "Match all query parameters given as arguments and keyword arguments" ["all_of"]
@portable_class
class all_of(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

#!defFunction(anvil.tables.query,_,*query_expressions)!2: "Match any query parameters given as arguments and keyword arguments" ["any_of"]
@portable_class
class any_of(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

#!defFunction(anvil.tables.query,_,*query_expressions)!2: "Match none of the query parameters given as arguments and keyword arguments" ["none_of"]
@portable_class
class none_of(object):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

#!defFunction(anvil.tables.query,_,*query_expressions)!2: "Match none of the query parameters given as arguments and keyword arguments" ["not_"]
not_ = none_of




import anvil

_config = None

def get_client_config():
    global _config
    if _config is None:
        _config = anvil._get_service_client_config("/runtime/services/anvil/users.yml")
    return _config



import anvil.google.auth
import anvil.regex
import anvil.server as rpc

add_missing_fields = True

def _list_gen(request_page_fn):
    queue = []
    start_index = 1
    total_results = None

    while True:

        if len(queue) > 0:
            yield queue.pop(0)
        elif total_results == None or start_index <= total_results:
            result = request_page_fn(start_index)

            total_results = result["total_results"]
            start_index += len(result["items"])

            if "items" in result and len(result["items"]) > 0:
                queue.extend(result["items"])
        else:
            break


def wrap_gen(items, item_class, creds):
    for i in items:
        yield item_class(i, creds)

class ApiItem(object):

    def __getitem__(self, item_name):

        if item_name in self._obj:
            return self._obj[item_name]
        else:
            raise AttributeError("Could not find item: %s" % item_name)

    def __getattr__(self, attr_name):
        if attr_name == "id":
            return self._obj["id"]
        else:
            raise AttributeError("Could not find attribute: %s" % attr_name)

    def __init__(self, dict, creds):
        self._obj = dict
        self.creds = creds

class Cell(ApiItem):

    def __init__(self, obj, creds):
        ApiItem.__init__(self, obj, creds)

        # These won't change, so set them here.

        self.row = self._obj["row"]
        self.col = self._obj["col"]


    def __getattr__(self, name):
        if name == "value":
            # Return the calculated value of the cell
            return self["value"]
        elif name == "input_value":
            # Return the input value of the cell
            return self["input_value"]

    def __setattr__(self, name, value):
        if name == "value" or name == "inputValue":
            url = self._obj["edit_url"]

            r = rpc.call("anvil.private.google.sheets.update_cell", self._obj["id"], url, self.row, self.col, value, self.creds)

            self._obj = r

            return value
        else:
            object.__setattr__(self, name, value)

    def __str__(self):
        return "<Google Worksheet Cell: %s>" % self._obj["value"]

class Worksheet(ApiItem):

    def __getattr__(self, name):
        if name == "fields":
            return self.get_fields()
        elif name == "rows":
            return list(self.list_rows())
        elif name == "cells":
            return list(self.list_cells())
        else:
            raise AttributeError("Could not find attribute: %s" % name)

    def __getitem__(self, cell_tuple):
        row, col = cell_tuple

        return self.get_cell(row, col)

    def get_fields(self):
        cs = self._obj["col_count"]

        return [anvil.regex.replace(c.value, "[^A-Za-z0-9\\-]", "").lower() for c in self.list_cells(1,1,1,cs)]

    def list_rows(self, limit=100, **kwargs):

        url = self._obj["list_feed_url"]

        query = ""
        if (len(kwargs) > 0):

            for k in kwargs.keys():
                if (len(query) > 0):
                    query += " and "
                if isinstance(kwargs[k], int) or isinstance(kwargs[k], float):
                  query += k + "=" + str(kwargs[k])
                else:
                  query += k + "=\"" + str(kwargs[k]) + "\""

        def next_page_fn(start_index):
          return rpc.call("anvil.private.google.sheets.list_rows", url, query, limit, self.creds, start_index=start_index)
        
        return _list_gen(next_page_fn)

    def add_row(self, **kwargs):

        values = {}
 
        count = 0
        for f in [k for k in kwargs if kwargs[k]!=""]:
            field_name = anvil.regex.replace(f, "[^A-Za-z0-9\\-]", "").lower()
            values[field_name] = str(kwargs[f])
            count += 1

        if count == 0:
            print("[WARNING: Skipping adding empty row to worksheet]")
            return None

        try:
          added_row = rpc.call("anvil.private.google.sheets.add_row", self._obj["list_feed_url"], values, self.creds)
        except Exception:
          added_row = None

        # The returned row will only contain fields that already exist in the sheet.
        # If we didn't specify any existing fields, no row will be returned at all.

        existing_fields = self.fields

        missing_fields = [k for k in values if not k in existing_fields]

        if len(missing_fields) > 0:

          if not add_missing_fields:
            if added_row:
              added_row.delete()
            raise Exception("Field(s) not found: %s" % ",".join(missing_fields))

          print("[WARNING: Adding missing fields to worksheet: %s]" % ",".join(missing_fields))

          # Create the field headers

          for i in range(len(missing_fields)):
            h = self.get_cell(1, len(existing_fields) + i + 1)
            h.value = missing_fields[i]



          if added_row:
            # If we previously added the row, we need to set the newly-added fields

            for f in missing_fields:
              added_row[f] = values[f]

          else:
            # Otherwise, we never added the row in the first place, 
            # so we just add it now directly.
            r = rpc.call("anvil.private.google.sheets.add_row", self._obj["list_feed_url"], values, self.creds)

            added_row = r

        return added_row

    def list_cells(self, min_row=None, max_row=None, min_col=None, max_col=None):
        url = self._obj["cells_feed_url"]

        query = ""

        if min_row:
            query += "&min-row=%d" % min_row
        if max_row:
            query += "&max-row=%d" % max_row
        if min_col:
            query += "&min-col=%d" % min_col
        if max_col:
            query += "&max-col=%d" % max_col


        def next_page_fn(start_index):
          return rpc.call("anvil.private.google.sheets.list_cells", url, query, self.creds, start_index=start_index)
        return wrap_gen(_list_gen(next_page_fn), Cell, self.creds)

    def get_cell(self, row, col):

        return Cell(rpc.call("anvil.private.google.sheets.get_cell", self._obj["cells_feed_url"], row, col, self.creds), self.creds)


    def __str__(self):
        return "<Google Worksheet: %s>" % self._obj["title"]


class Sheet(ApiItem):

    def __getitem__(self, name):
      if isinstance(name, str):
        for w in self.list_worksheets():
          if w._obj["title"] == name:
            return w
        raise KeyError("Spreadsheet contains no worksheet '%s'" % name)
      else:
        return list(self.list_worksheets())[name]

    def __getattr__(self, name):
        if name == "worksheets":
            return self
        else:
            raise AttributeError("Could not find attribute: %s" % name)

    def __iter__(self):
        return iter(self.list_worksheets())

    def __len__(self):
        return len(list(self.list_worksheets()))

    def list_worksheets(self):

        def request_page_fn(start_index):
            return rpc.call("anvil.private.google.sheets.list_worksheets", self._obj["worksheets_feed_url"], self.creds, start_index=start_index)

        return wrap_gen(_list_gen(request_page_fn), Worksheet, self.creds)


    def __str__(self):
        return "<Google Sheet: %s>" % self._obj["title"]




def login():
    return anvil.google.auth.login(['http://spreadsheets.google.com/feeds/'])

# This is so rarely used, and never documented, it was never ported to native modules.
#
# def list_sheets():
#     return wrap_gen(google.utils.feed_list_generator({
#         "url": "https://spreadsheets.google.com/feeds/spreadsheets/private/full",
#         "creds": "google-user"
#     }), Sheet)

def get_sheet(id, creds):
    return Sheet(rpc.call("anvil.private.google.sheets.get_sheet", id, creds), creds)


"""
id: drive_sheets
docs_url: /docs/integrations/google/google-drive#google-sheets
title: Google Sheets
description: |
  ```python
  db = app_files.my_sheet
  ```

  Google Sheets make ideal databases for Anvil apps. You can access sheets using Google Drive in the same way you access files.

  ```python
  ws = db["Sheet 1"]
  ```

  Once you have an object representing a Google Sheet, you can look up its individual worksheets. (It is also possible to look up a worksheet by number, eg `db[0]`, or obtain a list with `list(db.worksheets)`.)

  ```python
  print db["Sheet 1"].fields
  # Expect output similar to: '[name, age]'
  ```

  You should set up your Sheet using [Google Drive](https://drive.google.com) so that the column names are in the first row. Once your sheet is set up, you can get a list of the fields in a worksheet using the `fields` attribute.

  Fields will always be in lower case (even if your column headings use capital letters).

  ```python
  # Assuming columns "name" and "age"
  for r in db["Sheet 1"].rows:
    print "%s is %s years old" % (r["name"], r["age"])
  ```

  You can access rows of data like so:

  ```python
  # Convert age to dog years:
  for r in db["Sheet 1"].rows:
    r["age"] = 7 * int(r["age"])
  ```

  You can update them as well as reading them. You can assign numbers or strings,
  but the data you read from a row will always be a string:

  ```python
  ws = db["Sheet 1"]
  row = ws.add_row(name="Bob", age=56)
  ```

  You can add rows to your spreadsheet using the `add_row(...)` method:

  (Note: `add_row()` normally returns the new created row. However, if you
  attempt to insert an empty row, it will return `None`.)

  If you try to add a row containing a field that does not exist in the sheet, the field will be created automatically.
  You can change this behaviour by setting `anvil.google.sheets.add_missing_fields` to `False`, in which case an Exception will be raised if you try to set a non-existent field.

  ```python
  row.delete()
  ```

  You can delete a row using the `delete()` method.

  Beware: If you delete a row, other rows in the worksheet may shift unpredictably. If you delete a row, you should not re-use any other `row` objects from that worksheet. Call `list_rows()` again to find the rows you want again.

  ```python
  ## Display the calculated value of
  ## the cell in row 1, col 2

  print ws[1,2].value

  ## Display the formula in cell (1,2)

  print ws[1,2].input_value

  ## Set the value of a cell

  ws[3,4].value = "Some text"
  ```

  It is also possible to access individual cells of your spreadsheet.

"""



import collections, json, os, psutil, random, signal, subprocess, sys, threading, time, traceback, platform
from ws4py.client.threadedclient import WebSocketClient


# Configuration

TIMEOUT = int(os.environ.get("DOWNLINK_WORKER_TIMEOUT", "30"))
DROP_PRIVILEGES = os.environ.get("DROP_PRIVILEGES")
RUNTIME_ID = os.environ.get("RUNTIME_ID", None) or ('python2-full' if sys.version_info[0] < 3 else 'python3-full')
USER_ID = os.environ.get("DOWNLINK_USER_ID", None)
ORG_ID = os.environ.get("DOWNLINK_ORG_ID", None)
APP_CACHE_SIZE = int(os.environ.get("APP_CACHE_SIZE", "16"))
ENABLE_PDF_RENDER = os.environ.get("ENABLE_PDF_RENDER")
PER_WORKER_SOFT_MEMORY_LIMIT = int(os.environ["PER_WORKER_SOFT_MEMORY_LIMIT_MB"])*1024*1024 \
                                    if "PER_WORKER_SOFT_MEMORY_LIMIT_MB" in os.environ else None

IS_WINDOWS = "Windows" in platform.system() or "CYGWIN" in platform.system()

for V in ["DOWNLINK_WORKER_TIMEOUT", "DROP_PRIVILEGES", "RUNTIME_ID", "DOWNLINK_USER_ID", "DOWNLINK_ORG_ID", "APP_CACHE_SIZE", "ENABLE_PDF_RENDER"]:
    if V in os.environ:
        del os.environ[V]

# Worker modules register themselves here
workers_by_id = {}

# Cache app content
app_cache = collections.OrderedDict()


def send_with_header(json_data, blob=None):
    """"Send data to the API router"""
    connection.send_with_header(json_data, blob)

# Host state

launch_worker = None
launch_pdf_worker = None

connection = None

rnd = random.SystemRandom()
MY_SESSION_ID = "".join((rnd.choice("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789") for _ in range(20)))

draining_start_time = None

def maybe_quit_if_draining_and_done():
    if draining_start_time is not None and len(workers_by_id) == 0:
        if time.time() < draining_start_time + 10:
            print("Giving API 10 seconds' grace for drain...")
            def f():
                time.sleep(draining_start_time + 10 - time.time())
                maybe_quit_if_draining_and_done()
            threading.Thread(target=f).start()
        else:
            print("Drain complete. Exiting.")
            os._exit(0)


# Utility functions

def get_demote_fn(app_id):
    if os.name == "nt":
        return None

    # TODO: Use app_id here to seed UID generation. It might be an actual app ID, or None
    uid = 20000
    def do_demotion():
        if DROP_PRIVILEGES and os.getuid() == 0:
            os.setgroups([])
            os.setgid(uid)
            os.setegid(uid)
            os.setuid(uid)
            os.seteuid(uid)

        # Give ourselves an isolated process group so we can take child processes with us when we go
        os.setpgid(0, 0)

    return do_demotion


class PopenWithGroupKill(subprocess.Popen):
    def terminate(self):
        try:
            os.killpg(self.pid, 9)
        except:
            pass
        super(PopenWithGroupKill, self).terminate()


# Handle communication with API router
class Connection(WebSocketClient):
    def __init__(self, url, key):
        print("Connecting to " + url)
        WebSocketClient.__init__(self, url)

        self._sending_lock = threading.RLock()
        self._send_next_bin = None
        self._key = key

    def opened(self):
        print("Anvil websocket open")
        spec = {
            'runtime': RUNTIME_ID,
            'session_id': MY_SESSION_ID,
        }

        if USER_ID is not None:
            spec['user_id'] = USER_ID
        elif ORG_ID is not None:
            spec['org_id'] = ORG_ID

        id = os.environ.get("DOWNLINK_ID", None)
        if id:
            spec['id'] = id
        self.send(json.dumps({
            'key': self._key,
            'v': 2,
            'spec': spec,
        }))

    def closed(self, code, reason=None):
        print("Anvil websocket closed (code %s, reason=%s)" % (code, reason))
        # The world has ended. Let whatever is in charge of restarting us sort it out.
        os._exit(1)

    def received_message(self, message):
        try:
            self._received_message(message)
        except Exception as e:
            print("Error in received_message():")
            traceback.print_exc()
            raise

    def _received_message(self, message):

        if message.is_binary:
            self._send_next_bin(message.data)

        else:
            data = json.loads(message.data.decode())
            #print "Received: " + repr(data)

            type = data["type"] if 'type' in data else None
            id = data["id"] if 'id' in data else None

            if 'auth' in data:
                print("Downlink authenticated OK")

            elif 'output' in data:
                # Output from something this worker has called.
                calling_worker = workers_by_id.get(data.get('id'))
                originating_call = calling_worker.outbound_ids.get(id) if calling_worker is not None else None

                if originating_call is not None:
                    data['id'] = originating_call
                    self.send_with_header(data)
                else:
                    print("Bogus output, probably for an old request (worker: %s): %s" %
                          ("FOUND" if calling_worker else "MISSING", repr(data)[:100]))

            elif type in ["CALL_WITH_APP", "LAUNCH_BACKGROUND_WITH_APP", "CALL", "LAUNCH_BACKGROUND", "LAUNCH_REPL"]:

                if "app" not in data:
                    cached_app = app_cache.get((data["app-id"], data["app-version"]))
                    if cached_app is not None:
                        #print("Filling out app from cache for %s" % ((data["app-id"], data["app-version"]),))
                        data["type"] += "_WITH_APP"
                        data["app"] = cached_app

                #print "Launching new worker for ID " + id
                if draining_start_time:
                    self.send_with_header({"id": id, "error": {"type": "anvil.server.DownlinkDrainingError", "message": "New call routed to draining downlink"}})
                else:
                    if data.get("command", None) == "anvil.private.pdf.do_print":
                        if launch_pdf_worker:
                            launch_pdf_worker(data)
                        else:
                            self.send_with_header({"id": id, "error": {"type": "anvil.server.RuntimeUnavailableError", "message": "PDF Rendering unavailable"}})
                    else:
                        launch_worker(data)

                #print "Launched"

            elif type in ["REPL_COMMAND", "REPL_KEEPALIVE", "TERMINATE_REPL"]:
                worker = workers_by_id.get(data['repl'])

                # TODO allow REPL commands to be run on us too

                if worker is not None:
                    worker.handle_inbound_message(data)
                else:
                    print("Couldn't find repl %s; current workers: %s" % (data['repl'], workers_by_id.keys()))
                    connection.send_with_header(
                        {'error': {'type': 'anvil.server.NotRunningTask', 'message': 'No such REPL running'},
                         'id': data['id']}
                    )

            elif type == "KILL_TASK":

                worker = workers_by_id.get(data['task'])
                if worker is not None:
                    worker.kill_background_task()

            elif type == "GET_TASK_STATE":

                worker = workers_by_id.get(data['task'])
                if worker is not None:
                    worker.get_task_state(data)
                else:
                    connection.send_with_header(
                        {'error': {'type': 'anvil.server.NotRunningTask', 'message': 'No such task running'},
                         'id': data['id']})

            elif type == "CHUNK_HEADER":
                if data['requestId'] in workers_by_id:
                    worker = workers_by_id[data['requestId']]

                    def send_next_bin(bin_data):
                        worker.handle_inbound_message(data, bin_data)
                        self._send_next_bin = None

                    self._send_next_bin = send_next_bin
                else:
                    print("Ignoring media for unknown request %s" % data['requestId'])
                    self._send_next_bin = lambda x: 0

            elif (type is None or type == "PROVIDE_APP") and "id" in data:
                if type == "PROVIDE_APP":
                    #print("PROVIDE_APP: Cache fill for %s" % ((data["app-id"], data["app-version"]),))
                    app_cache[(data["app-id"], data["app-version"])] = data["app"]
                    if len(app_cache) > APP_CACHE_SIZE:
                        app_cache.popitem(False)

                if id in workers_by_id:
                    workers_by_id[id].handle_inbound_message(data)
                elif id.startswith("downlink-keepalive"):
                    pass # We don't care about these
                else:
                    print("Bogus reply: " + repr(data)[:100])

            elif type is None and "error" in data:
                print("Fatal error from Anvil server: " + str(data["error"]))
                os._exit(1)
            else:
                print("Anvil websocket got unrecognised message: "+repr(data))

    def send(self, payload, binary=False):
        with self._sending_lock:
            return WebSocketClient.send(self, payload, binary)

    def send_with_header(self, json_data, blob=None):
        with self._sending_lock:
            WebSocketClient.send(self, json.dumps(json_data), False)
            if blob is not None:
                WebSocketClient.send(self, blob, True)


# Defined in two places, so it can be used by BaseWorker and the full-python worker. Yeuch.
def report_worker_stats(self):
    p = self.proc_info
    if p is None:
        return {}
    try:
        cpu = p.cpu_times()
        mem = p.memory_full_info()
        return {
            "info": self.task_info,
            "age":  time.time() - p.create_time(),
            "cpu": {
                "user": cpu.user + cpu.children_user,
                "system": cpu.system + cpu.children_system,
                "total": cpu.user + cpu.system + cpu.children_user + cpu.children_system
            },
            "mem": {"vms": mem.vms, "uss": mem.uss},
        }
    except psutil.Error:
        return {}


# Shared tools for managing worker processes.
# Nomenclature: "Inbound" calls come from the API server. "Outbound" calls come from the server.
class BaseWorker(object):
    def __init__(self, initial_msg, task_info):
        self.req_ids = set()
        self.outbound_ids = {} # Outbound ID -> inbound ID it came from
        self._media_tracking = {} # reqID -> (set([mediaId, mediaId, ]), finishedCallback)
        self.lock = threading.RLock() # TODO do we need this?
        self.start_times = {}
        self.proc_info = None
        self.task_info = task_info

        self.initial_req_id = initial_msg['id']

    # Handle bookkeeping for which requests we're handling and waiting for

    def record_outbound_call_started(self, outbound_msg):
        outbound_id = outbound_msg['id']
        if outbound_id in workers_by_id:
            raise Exception("Duplicate ID: %s" % outbound_id)
        self.outbound_ids[outbound_msg['id']] = outbound_msg.get('originating-call', self.initial_req_id)
        workers_by_id[outbound_id] = self

    def record_outbound_call_complete(self, outbound_id):
        self.outbound_ids.pop(outbound_id, None)
        workers_by_id.pop(outbound_id, None)

    def record_inbound_call_started(self, inbound_msg):
        inbound_id = inbound_msg['id']
        self.req_ids.add(inbound_id)
        self.start_times[inbound_id] = time.time()
        workers_by_id[inbound_id] = self

    def record_inbound_call_complete(self, inbound_id):
        self.req_ids.discard(inbound_id)
        self.start_times.pop(inbound_id, None)
        workers_by_id.pop(inbound_id, None)

        if len(self.req_ids) == 0:
            self.on_all_inbound_calls_complete()

        maybe_quit_if_draining_and_done()

    def clean_up_all_outstanding_records(self):
        for id in self.req_ids:
            self._media_tracking.pop(id, None)
            workers_by_id.pop(id, None)
        for id in self.outbound_ids:
            self._media_tracking.pop(id, None)
            workers_by_id.pop(id, None)

    def ensure_id_is_mine(self, req_id):
        if not (req_id in self.req_ids or req_id in self.outbound_ids):
            raise Exception("Worker attempted to send an ID that doesn't belong to it")

    # Events to be overridden by children

    def handle_inbound_message(self, msg, bindata=None):
        raise Exception("handle_inbound_message() not implemented")

    def on_all_inbound_calls_complete(self):
        raise Exception("on_all_inbound_calls_complete() not implemented")

    def repl_keepalive(self):
        raise Exception("repl_keepalive() not implemented")

    # A common task is to track when the worker has finished sending media for a particular request,
    # so we can safely kill it.

    def on_media_complete(self, msg, callback):
        """Register a callback to execute when the worker has finished sending all the media in the given message."""
        media_ids = set()
        for o in msg.get("objects", []):
            if "DataMedia" in o.get("type", []):
                media_ids.add(o["id"])
        if len(media_ids) == 0:
            callback()
        else:
            # print("Waiting for media for request '%s': %s" % (msg['id'], repr(list(media_ids))))
            self._media_tracking[msg['id']] = (media_ids, callback)

    def transmitted_media(self, request_id, media_id):
        """The worker has finished sending the specified media object; call any necessary callbacks"""

        # print("Media complete: '%s', '%s'" % (request_id, media_id))
        if request_id in self._media_tracking:
            media_ids, callback = self._media_tracking[request_id]
            media_ids.discard(media_id)
            if len(media_ids) == 0:
                callback()
                del self._media_tracking[request_id]

    # Slightly awkward shimming of profiling information into a response message
    def fill_out_profiling(self, response_msg, description="Downlink dispatch"):
        """Add profiling information to a response message"""

        p = response_msg.get("profile", None)
        response_msg["profile"] = {
            "origin": "Server (Python)",
            "description": description,
            "start-time": float(self.start_times.get(response_msg['id'], 0)*1000),
            "end-time": float(time.time()*1000)
        }
        if p is not None:
            response_msg["profile"]["children"] = [p]

            for o in response_msg.get("objects", []):
                if o["path"][0] == "profile":
                    o["path"].insert(1,"children")
                    o["path"].insert(2, 0)

    report_stats = report_worker_stats


# Import the actual worker modules

def init_pdf_worker():
    global launch_pdf_worker

    if sys.version_info < (3,7,0):
        print("Warning: PDF Rendering requires Python 3.7. Renderer not initialised")
    elif IS_WINDOWS:
        print("Warning: PDF Rendering not supported on Windows. Renderer not initialised")
    else:
        from . import pdf_renderer
        launch_pdf_worker = pdf_renderer.launch


if RUNTIME_ID == "pdf-renderer":
    init_pdf_worker()
elif RUNTIME_ID.endswith('-sandbox'):
    from . import pypy_sandbox
    launch_worker = pypy_sandbox.launch
else:
    from . import full_python
    launch_worker = full_python.launch

    if ENABLE_PDF_RENDER:
        init_pdf_worker()


def signal_drain(_signum, _frame):
    global draining_start_time
    connection.send_with_header({
        "type": "DRAIN"
    })
    print("Draining downlink. %s call(s) remaining:" % len(workers_by_id))
    print(list(workers_by_id.keys()))
    draining_start_time = time.time()
    maybe_quit_if_draining_and_done()


def report_stats():
    workers = set(workers_by_id.values())
    worker_stats = []
    for worker in workers:
        stats = worker.report_stats()
        mem_usage = stats.get("mem", {}).get("uss", 0)
        if PER_WORKER_SOFT_MEMORY_LIMIT is not None and mem_usage > PER_WORKER_SOFT_MEMORY_LIMIT:
            print("Worker is using %.0fMB: %s " % (mem_usage/(1024*1024.0), stats.get('info')))
            worker.drain()
        worker_stats.append(stats)
    connection.send_with_header({
        "type": "STATS",
        "data": worker_stats
    })



def run_downlink_host():
    global connection

    url = os.environ.get("DOWNLINK_SERVER", "ws://127.0.0.1:3000/downlink")
    key = os.environ.get("DOWNLINK_KEY", "ZeXiedeaceimahm1ePhaguvu5Ush9E")
    os.environ['TZ'] = 'UTC'

    for v in ["DOWNLINK_SERVER", "DOWNLINK_KEY"]:
        if v in os.environ:
            del os.environ[v]

    connection = Connection(url, key)

    connection.connect()

    if not IS_WINDOWS:
        try:
            signal.signal(signal.SIGUSR2, signal_drain)
        except Exception as e:
            print("Failed to add signal handler: %s" % e)

    n = 0
    while True:
        try:
            for _ in range(6):
                time.sleep(5)
                report_stats()

            connection.send_with_header({
                "type": "CALL",
                "id": "downlink-keepalive-%d" % n,
                "command": "anvil.private.echo",
                "args": ["keep-alive"],
                "kwargs": {},
            })
            n += 1
        except Exception as e:
            print("Keepalive failed. The downlink has probably disconnected.")
            print(e)
            os._exit(1)



import datetime
import time


class tzoffset(datetime.tzinfo):

	#!defMethod(_,[seconds], [minutes], [hours])!2: "Create a timezone with a specific offset. Use an offset in seconds, minutes or hours" ["__init__"]
    def __init__(self, **kwargs):
        # Must initialise with seconds OR minutes OR hours, or with no arguments to get an offset of zero.
        if len(kwargs) > 1 or (len(kwargs) == 1 and not 'seconds' in kwargs and not 'hours' in kwargs and not 'minutes' in kwargs):
            raise Exception("tzoffset must be initialised with precisely one of 'seconds', 'minutes' or 'hours' keyword arguments")

        self._offset = datetime.timedelta(**kwargs)


    def utcoffset(self, dt):
        return self._offset

    def dst(self, dt):
        return datetime.timedelta()

    def tzname(self, dt):
        return None

    def __repr__(self):
        return "tzoffset(%s hours)" % (self._offset.total_seconds() / 3600)
#!defClass(anvil.tz, tzoffset)!:


# People should probably never use this, but we provide it for completeness.
class tzlocal(tzoffset):

	#!defMethod(_)!2: "Use the local timezone of the browser" ["__init__"]
    def __init__(self):
        if time.daylight:
            s = -time.altzone
        else:
            s = -time.timezone
        tzoffset.__init__(self, seconds=s)


    def __repr__(self):
        return "tzlocal(%s hour offset)" % (self._offset.total_seconds() / 3600)
#!defClass(anvil.tz, tzlocal, anvil.tz.tzoffset)!:


class tzutc(tzoffset):

	#!defMethod(_)!2: "Create a timezone set to utc" ["__init__"]
    def __init__(self):
        tzoffset.__init__(self, minutes=0)

    def __repr__(self):
        return "tzutc"
#!defClass(anvil.tz, tzutc, anvil.tz.tzoffset)!:

#!defModuleAttr(anvil.tz)!1: {name: 'UTC', type: 'any', description: 'An object representing the UTC timezone'}
UTC = tzutc()



import anvil.google.auth
import anvil.google.sheets
import anvil
import anvil.server as rpc

google_api_scopes = ["https://www.googleapis.com/auth/drive", 'http://spreadsheets.google.com/feeds/']

try:
    basestring
except NameError:
    basestring = str

#####
# Internal helper methods
#####

def _query_single(parent_id, creds, **kwargs):
    r = rpc.call("anvil.private.google.drive.list_files", parent_id, creds, max_results=1, **kwargs)

    items = r["items"]
    if len(items) == 0:
        return None
    else:
        return items[0]

def _list_gen(request_page_fn):
    queue = []
    nextPageToken = ""

    while True:

        if len(queue) > 0:
            yield queue.pop(0)
        elif nextPageToken or nextPageToken == "":

            result = request_page_fn(nextPageToken)

            if 'nextPageToken' in result:
                nextPageToken = result["nextPageToken"]
            else:
                nextPageToken = None

            if "items" in result and len(result["items"]) > 0:
                queue.extend(result["items"])
                yield queue.pop(0)
            else:
                break
        else:
            break



def wrap_gen(items, creds, item_class=None):
    for i in items:
        if item_class is None:
          yield wrap_item(i, creds)
        else:
          yield item_class(i, creds)

def wrap_item(f, creds):
    if f == None:
        return None
    elif f["mimeType"] == 'application/vnd.google-apps.folder':
        return Folder(f, creds)
    elif f["mimeType"] == 'application/vnd.google-apps.spreadsheet':
        return anvil.google.sheets.get_sheet(f["id"], creds)
    else:
        return File(f, creds)

def create_item_simple(title, folder, mime_type, creds):

    # Use simple metadata-only endpoint to create item.

    return wrap_item(rpc.call("anvil.private.google.drive.create_item_simple", title, folder.id, mime_type, creds), creds)

def create_item_multipart(title, folder, content, creds, content_type=None):

    return wrap_item(rpc.call("anvil.private.google.drive.create_item_multipart", title, folder.id, content, creds, content_type=content_type), creds)

def update_metadata(item, metadata, creds):

    ctor = type(item)
    return ctor(rpc.call("anvil.private.google.drive.update_metadata", item.id, metadata, creds), creds)


#####
# Google Drive classes
#####

class ApiItem(object):

    def __getitem__(self, item_name):

        if item_name in self._obj:
            return self._obj[item_name]
        else:
            raise AttributeError("Could not find item: %s" % item_name)

    def __getattr__(self, attr_name):
        if attr_name == "id":
            return self._obj["id"]
        else:
            raise AttributeError("Could not find attribute: %s" % attr_name)

    def __init__(self, dict, creds):
        self._obj = dict
        self.creds = creds



class DriveItem(ApiItem):

    def __setitem__(self, item_name, value):
        if item_name in ["title"]: # List of valid fields to set
            update_metadata(self, {item_name: value}, self.creds)
        else:
            object.__setitem__(self, item_name, value)

    def move(self, dest_folder):
        return update_metadata(self, { 'parents': [{ 'id': dest_folder.id, 'old_id': self['parents'][0]['id'] }]}, self.creds)

    def trash(self):
        ctor = type(self)
        return ctor(rpc.call("anvil.private.google.drive.trash", self.id, self.creds), self.creds)

    def delete(self):
        rpc.call("anvil.private.google.drive.delete", self.id, self.creds)


class File(DriveItem,rpc.LazyMedia):

    def __init__(self,d,creds):
      rpc.LazyMedia.__init__(self,d["anvil_LazyMedia"])
      ApiItem.__init__(self,d,creds)

    def __getattribute__(self, attr_name):
        # Using __getattribute__ instead of __getattr__ here is a massive hack,
        # because client-side LazyMedia defines __getattribute__ as a workaround for some
        # Skulpt nastiness. Wheels within wheels...
        try:
            return object.__getattribute__(self, attr_name)
        except AttributeError:
            if attr_name == "revisions":
                return list(self.list_revisions())
            elif attr_name == "url" or attr_name == "content_type" or attr_name == "length" or attr_name == "name":
                return anvil.Media.__getattribute__(self, attr_name)
            else:
                return DriveItem.__getattr__(self, attr_name)

    def __setattr__(self, attr_name, value):
        if attr_name == "content_type":
            raise Exception("Cannot set a Google Drive file's content type as an attribute. Use set_media() instead.")
        # Was "elif attr_name == "_obj" or attr_name == "creds":"
        # but that broke at least _spec, and possibly something else
        else:
            object.__setattr__(self, attr_name, value)
        #else:
        #    raise Exception("Cannot set attribute '" + attr_name + "' of a Google Drive file")


    def get_bytes(self, revision=None):

        if revision:
            url = revision["downloadUrl"]
        else:
            url = self._obj["downloadUrl"]

        b = rpc.call("anvil.private.google.drive.get_content", url, self.creds).get_bytes()
        self._obj["fileSize"] = len(b)
        return b

    def get_content_type(self):
        return self["mimeType"]

    def get_name(self):
        return self["title"]

    def get_length(self):
        return int(self["fileSize"])

    def set_bytes(self, content):

        # Use simple upload to replace content. Don't need to touch metadata.

        content_type = "text/plain"

        if not isinstance(content, str):
            raise Exception("The 'bytes' of a file must be a Python string. To upload a Media object, use set_media().")

        r = rpc.call("anvil.private.google.drive.set_content", self.id, content, self.creds)

        self._obj["mimeType"] = content_type
        return File(r, self.creds)

    def set_media(self, media):
        if not isinstance(media, anvil.Media):
            raise Exception("set_media() must be called with a Media object.")

        r = rpc.call("anvil.private.google.drive.set_content", self.id, media, self.creds)

        self._obj["mimeType"] = media.content_type
        return File(r, self.creds)


    def list_revisions(self):

        def request_page_fn(page_token):
            return rpc.call("anvil.private.google.drive.list_file_revisions", self.id, self.creds, page_token)

        return wrap_gen(_list_gen(request_page_fn), self.creds, FileRevision);

    def __str__(self):
        return "<Google Drive File: %s>" % self["title"]


class FileRevision(ApiItem):

    def __getattr__(self, attr_name):
        if attr_name == "date":
            return self["modifiedDate"]
        else:
            return ApiItem.__getattr__(self, attr_name)


class Folder(DriveItem):

    def __getattr__(self, attr_name):
        if attr_name == "files":
            return list(self.list_files())
        elif attr_name == "folders":
            return list(self.list_folders())
        else:
            return DriveItem.__getattr__(self, attr_name)

    # Bear in mind that these query strings must correspond exactly to the server
    # proxy whitelist, or the request will be refused for app files.

    def list_files(self):

        def request_page_fn(page_token):
            return rpc.call("anvil.private.google.drive.list_files", 
                            self.id, 
                            self.creds, 
                            page_token=page_token,
                            trashed=False,
                            mime_type='!application/vnd.google-apps.folder')

        return wrap_gen(_list_gen(request_page_fn), self.creds);

    def list_folders(self):

        def request_page_fn(page_token):
            return rpc.call("anvil.private.google.drive.list_files", 
                            self.id, 
                            self.creds, 
                            page_token=page_token,
                            trashed=False,
                            mime_type='application/vnd.google-apps.folder')

        return wrap_gen(_list_gen(request_page_fn), self.creds);

    def get(self, title):
        return wrap_item(_query_single(self.id, self.creds, title=title, trashed=False), self.creds)

    def get_by_id(self, id):
        if not isinstance(id, basestring):
            raise Exception("ID must be a string")

        if id == "":
            raise Exception("The empty string is not a valid ID")

        return wrap_item(rpc.call("anvil.private.google.drive.get_file_by_id", self.id, id, self.creds), self.creds)

    def create_folder(self, title):
        return create_item_simple(title, self, "application/vnd.google-apps.folder", self.creds)

    def create_file(self, title, content_bytes = None, content_type = "text/plain"):
        if content_bytes:
            return create_item_multipart(title, self, content_bytes, self.creds, content_type=content_type)
        else:
            return create_item_simple(title, self, content_type, self.creds)

    def __str__(self):
        return "<Google Drive Folder: %s>" % self["title"]

#####
# Documentation
#####
"""
id: google_drive
docs_url: /docs/integrations/google/google-drive
title: Google Drive
description: |
  <div class="tutorial-link">Tutorial: Photo gallery with Google Drive<br><a href="/blog/photo-gallery"><i class="fa fa-play"></i> Play video</a></div>

  This service allows you to integrate Google Drive functionality into your Anvil app.
  You can select "app files" to which all users have access, or ask users to log in and
  then create, update and delete files in their own Google Drives.

  To add the Google Drive service to your app, click the plus sign (<i style="color:#428bca" class="fa fa-plus"></i>)
  next to **Services** in the [App browser](#app_browser).

  <img src="img/add_service.png" style="border: 1px solid #ccc; margin: 10px 0;">

includes: [drive_app_files,drive_file_io,drive_file_folders,drive_file_management,drive_sheets,drive_user_files]
"""

"""
id: drive_app_files
docs_url: /docs/integrations/google/google-drive#drive-app-files
title: App files
tooltip: Click for more about app files
description: |
  ```python
  from anvil.google.drive import app_files

  f = app_files.hello_txt
  print "File contents: " + f.get_bytes()
  ```

  App files are files or folders from your Google Drive that are available to any user of the app. Your users do not have to log in with Google; these files can always be accessed by this app.

  From the Google configuration page (under **services** in the [App browser](#app_browser)), you can click **Add app files** to add a file or folder from your Google Drive.

  Each app file (or folder) has a Python identifier derived from its filename.
  You can access these files as <code>app_files.<i>&lt;python-identifier&gt;</i></code>.

popup_includes: [drive_app_files_extra_link]
includes: [drive_permissions]
"""

"""
id: drive_permissions
docs_url: /docs/integrations/google/google-drive#drive-app-files
title: Permissions
tooltip: Click for more about permissions
description: |
  App files can have three levels of permission, indicated by symbols in the Google configuration page:

  <i class="fa fa-pencil"></i> **Client can read and write**<br>
  By default, any code in your app can read or modify an app file.
  This is convenient, but it also means anyone who can access your app could read or modify that file.
  (This is because the code in your forms - the *client code* - runs in the user's browser, so a malicious person could change that code
  to do whatever they like.) This is fine if you only share the app with people you trust - or if the app file
  is something you deliberately want the world to have access to.

  <i class="fa fa-eye"></i> **Client can read**<br>
  You can also make app files "read-only" for clients. This means that
  all of your code can read this app file, but only your [server modules](#server_modules) can change it.

  <i class="fa fa-eye-slash"></i> **No client access**<br>
  Finally, you can make your app file entirely private. This means only
  your [server modules](#server_modules) can read or write it.
"""


"""
id: drive_app_files_extra_link
docs_url: /docs/integrations/google/google-drive#drive-app-files
description:
 - "[Learn more about using Google Drive](#google_drive)"

"""

"""
id: drive_file_folders
docs_url: /docs/integrations/google/google-drive#folders
title: Folders
description: |
  ```python
  from anvil.google.drive import app_files

  folder = app_files.my_folder

  for f in folder.list_files():
    print f["title"]
  ```

  If you have an object representing a Google Drive folder, you can list the files in that folder with `list_files()`.

  `list_files()` returns an iterator, so you can loop over it in a `for` loop.

  ```python
  l1 = folder.files
  l2 = folder.folders
  print "This folder has %s files and %s folders in it" % (len(l1), len(l2))
  ```

  If you want to do more than just loop through the files, you can request a full list of files directly. Bear in mind that this might be slow if there are many files in the folder.

  Remember, a folder can contain other folders as well as files.

  ```python
  my_file = folder.get("my_file.txt")
  my_folder = folder.get("My Subfolder")
  ```

  You can even get an item by its title:

  ```python
  f = folder.create_file("new_file.txt")

  id = f.id

  # ... later ...

  my_file = folder.get_by_id(id)
  ```

  Every file has an ID, and you can get a file from a folder by its ID too,
  with `get_by_id`

"""
"""
id: drive_file_io
docs_url: /docs/integrations/google/google-drive#file-io
title: Files
description: |
  ```python
  from anvil.google.drive import app_files

  f = app_files.hello_txt
  f.set_bytes("My name is Bob.")
  ```

  If you have an object representing a Google Drive file, you can get or set its contents as a string with the `get_bytes()` and `set_bytes()` functions.

  ```python
  from anvil.google.drive import app_files

  f = app_files.my_file
  f.set_media(self.file_loader_1.file)
  ```

  You can also upload a [Media](#media) object (for example, from a [FileLoader](#fileloader) component) to a Google File.

  ```python
  from anvil.google.drive import app_files

  f = app_files.my_image

  self.image_1.source = f
  ```

  You can use a Google Drive file as a [Media](#media) object. Here we use a Google Drive file as the source of an [Image](#image) component:

  ```python
  from anvil.google.drive import app_files

  folder = app_files.my_folder

  new_file = folder.create_file("new_file.txt")
  ```

  To create files, use the `Folder.create_file` method.

  Calling `create_file` with a single file name argument creates a
  new file with MIME type `text/plain`.

  ```python
  new_file = folder.create_file("new_file.txt",
                                "Hello, world!")

  new_image = folder.create_file("new_file.jpg",
                                 file_loader_1.file)
  ```

  You can also pass an optional second argument, providing the initial
  content of the file. This can either be a string, or a [Media object](#media).
  If it is a Media object, the new file is automatically created with the correct
  MIME type.

  If possible, you should provide initial content when creating a file, rather than
  creating an empty file and then uploading its content. This way, if something goes
  wrong during upload then you won't end up with a new empty file.
"""
"""
id: drive_file_management
docs_url: /docs/integrations/google/google-drive#file-management
title: File management
description: |
  ```python
  from anvil.google.drive import app_files

  my_file = app_files.my_file
  folder = app_files.my_folder

  my_file.move(folder)
  ```

  You can move a Drive item (file or folder) from one folder to another by calling <code>move(<i>&lt;destination-folder&gt;</i>)</code>.

  ```python
  new_file = folder.create_file("new_file.txt")
  new_folder = folder.create_folder("new_folder")
  ```

  You can create a new file or folder, by calling <code>create_file(<i>&lt;title&gt;</i>)</code>
  or <code>create_folder(<i>&lt;title&gt;</i>)</code> on a folder.

  ```python
  file1 = folder.create_file("new_file.txt")
  file2 = folder.create_file("new_file.txt")

  saved_id = file1.id

  # (... some time later ...)

  # This will retrieve file1, not file2
  f = folder.get_by_id(saved_id)

  ```

  Google Drive permits you to create many files or folders with the same title. You can tell them apart by their `id` property,
  which is a unique string. If you want to store a reference to a file, you should use the `id`.

  ```python
  new_file.trash()
  new_folder.delete()
  ```

  You can put a Drive item in the trash by calling `trash()`, or delete it forever by calling `delete()` (this cannot be undone).
"""
"""
id: drive_user_files
docs_url: /docs/integrations/google/google-drive#user-files
title: Using a logged-in user's files
description: |
  As well as app files, you can use the Google service to access your users' own files when they log into your app with Google.

  To do this, you will need a Google API client ID (same as for the [Google REST API](#google_rest_api)).

  ```python
  import anvil.google.drive

  anvil.google.drive.login()

  folder = anvil.google.drive.get_user_files()

  for f in folder.list_files():
    print f["title"]
  ```

  To read and write files in a user's Google Drive, you must first call `anvil.google.drive.login()`, which will ask the user for permission to access their files. This is different from calling
  `anvil.google.auth.login()`, which only asks for their email address.

  Once a user has logged into your app using `anvil.google.drive.login()` you can read and write files in their Google Drive. (You can also pass a list of extra scopes to `anvil.google.drive.login()`, just like `anvil.google.auth.login()`.)

  You can get the top-level folder containing all the files in their drive by calling `anvil.google.drive.get_user_files()`.

  You will need to get a client ID and secret from the
  <a href="/doc/#linking_anvil_and_google" target="_blank">Google Developer Console <i class="fa fa-external-link"></i></a>
  if you want to use `anvil.google.drive.login()`. This client ID and secret must be pasted
  into the settings page for the Google service.

"""


#####
# API methods
#####

def login(extra_scopes=[]):
    return anvil.google.auth.login(google_api_scopes + extra_scopes)

def get_user_files():
    return Folder(rpc.call("anvil.private.google.drive.get_user_files"), "google-user")

class AppFilesCollection:

    def __getattr__(self, python_name):
        id = None
        for df in anvil.google.get_config().get("app_files", []):
            if df["python_name"] == python_name:
                id = df["id"]
                break

        if id == None:
            raise Exception("No such app file: %s" % python_name)

        f = rpc.call("anvil.private.google.drive.get_app_file", id)

        return wrap_item(f, "google-delegated")

app_files = AppFilesCollection()



__author__ = 'meredydd'

import random, string

import anvil
from . import _server



def _gen_id():
    return ''.join(random.SystemRandom().choice(string.ascii_letters + string.digits) for _ in range(10))


# requestId->_IncomingRequest
_incoming_requests = {}


class StreamingMedia(anvil.Media):
    def __init__(self, content_type, name):
        self._content_type = content_type
        self._content = b''
        self._incoming_content = []
        self._complete = False
        self._name = name

    def add_content(self, data, last_chunk=False):
        self._incoming_content.append(data)
        if last_chunk:
            self._content = b''.join(self._incoming_content)
            self._incoming_content = []
            self._complete = True

    def is_complete(self):
        return self._complete

    def get_content_type(self):
        return self._content_type

    def get_bytes(self):
        return self._content

    def get_url(self):
        raise None

    def get_name(self):
        return self._name


class IncomingReqResp:
    def __init__(self, json):
        self.media = {}
        self.capabilities = []

        def reconstruct_data_media(d):
            reconstructed = StreamingMedia(d['mime-type'], d.get("name", None))
            self.media[d['id']] = reconstructed
            return reconstructed

        self.json = _server._reconstruct_objects(json, reconstruct_data_media,
                                                 hold_back_value_types=True, collect_capabilities=self.capabilities)

        _incoming_requests[self.json["id"]] = self
        self.maybe_execute()

    def reconstruct_remaining_data(self):
        """Call this when you're ready to execute user code"""

        def assert_no_media(_):
            raise Exception("We shouldn't have any Media left by this point")

        self.json = _server._reconstruct_objects(self.json, assert_no_media)

    def add_binary_data(self, json, data):
        self.media[json['mediaId']].add_content(data, json['lastChunk'])
        if json['lastChunk']:
            self.maybe_execute()

    def is_ready(self):
        for id in self.media:
            if not self.media[id].is_complete():
                return False

        if holding_reqresps:
            return False

        return True

    def maybe_execute(self):
        if not self.is_ready():
            return

        del _incoming_requests[self.json["id"]]

        self.execute()



_next_hdr = None


def process_blob_header(hdr):
    global _next_hdr
    _next_hdr = hdr


def process_blob(blob):
    global _next_hdr
    _incoming_requests[_next_hdr['requestId']].add_binary_data(_next_hdr, blob)
    _next_hdr = None


# Machinery to suspend execution of all reqresps until an external event (app loading)
holding_reqresps = False


def release_reqresps():
    global holding_reqresps
    holding_reqresps = False
    for reqresp in list(_incoming_requests.values()):
        reqresp.maybe_execute()


def serialise(reqresp, do_send, collect_capabilities=None):
    media = []

    def enqueue_media(m):
        media_id = _gen_id()
        media.append((media_id, m))
        return {"id": media_id}

    reqresp = _server.fill_out_media(reqresp, enqueue_media, collect_capabilities=collect_capabilities)

    do_send(reqresp)

    for (id,m) in media:
        data = m.get_bytes()
        l = len(data)
        i = 0
        n = 0
        sent_once = False
        while i < l or not sent_once:
            chunk_len = min(l - i, 65536)

            do_send({'type': 'CHUNK_HEADER', 'requestId': reqresp['id'], 'mediaId': id,
                     'chunkIndex': n, 'lastChunk': (i + chunk_len == l)},
                    data[i:(i+chunk_len)])

            i += chunk_len
            n += 1
            sent_once = True



import json

from anvil_downlink_host import send_with_header as send_to_server
import anvil_downlink_host
from random import SystemRandom
import subprocess
import pychrome
import base64
import sys
import threading
import time
from select import select
import os.path
from pprint import pprint
from numbers import Number
from tempfile import NamedTemporaryFile, TemporaryDirectory

if sys.version_info < (3,0,0):
    import urllib
    _urlencode = lambda s: urllib.quote(s,safe='')
else:
    import urllib.parse
    _urlencode = lambda s: urllib.parse.quote(s,safe='')

if os.getuid() == 0:
    CHROME_SANDBOX_ARGS = ["sudo", "-u", "nobody"]
else:
    CHROME_SANDBOX_ARGS = []


# Sandboxing args for PS2PDF
def run_ps2pdf(quality, infile, outfile):
    with TemporaryDirectory() as tmpdir:
        if os.getuid() != 0:
            args = ["/usr/bin/ps2pdf", f"-dPDFSETTINGS=/{quality}", infile, outfile]
        else:
            os.mkdir(tmpdir+"/tmp")
            print("ls "+tmpdir)
            os.system("ls "+tmpdir)
            os.chmod(infile, 0o644)
            os.chmod(outfile, 0o666)
            os.chmod(tmpdir, 0o755)
            args = ["minijail0",
                    ## Errors on stdout, please (defaults to syslog)
                    "--logging=stderr",
                    ## Uncomment to catch minijail error logs that don't go to stderr despite the
                    ## previous line (grr).
                    ## NB if you do this in prod, you'll need to volume in /dev/log or minijail will
                    ## bail out
                    #"-b", "/dev/log",
                    # Run as 'nobody'
                    "-u", "nobody",
                    # Disable SysV shared-mem/IPC
                    "-l",
                    # Isolate from other processes
                    "-p",
                    # No hostname changes
                    "--uts",
                    # Freeze cgroups settings
                    "-N",
                    # Pivot-root to our temporary directory
                    "-P", tmpdir,
                    # Add a /tmp filesystem
                    "-t",
                    # Mount in system files ps2pdf needs
                    "-b", "/usr", "-b", "/bin", "-b", "/lib", "-b", "/lib64", "-b", "/etc", "-b", "/dev/urandom",
                    # Only permit a few system calls
                    "-S", anvil_downlink_host.__path__[0]+"/ps2pdf-seccomp.policy",
                    # Mount in the files (only the output is writable)
                    "-b", infile+",/input.pdf",
                    "-b", outfile+",/output.pdf,1",
                    # Now the actual invocation
                    "/usr/bin/ps2pdf",
                    f"-dPDFSETTINGS=/{quality}",
                    "/input.pdf",
                    "/output.pdf"
                    ]
        print(" ".join(args))
        return subprocess.run(args, capture_output=True)



# A generator that reads lines from a BufferedReader until the timeout expires. We can't use readline because that can block forever.
def non_blocking_readlines(f, timeout):
    start_time = time.time()
    byte_buffer = b""
    while time.time() < start_time + timeout:
        (readable,_,_) = select([f], [], [], 0.1)
        if f in readable:
            byte_buffer += f.read1(1024)
            lines = byte_buffer.split(b'\n')
            byte_buffer = lines[-1]
            for line in lines[:-1]:
                yield line.decode("utf-8")


# Returns True if we should wait for this network request to complete before printing.
def is_interesting_network_request(request):
    # This might be useful at some point, if we want to be cleverer about which requests to wait for
    # pprint({request['request']['url']: (request['initiator']['type'],request['initiator'].get('stack', None)))
    return True


# Yes, Chrome uses inches. Anvil uses sensible units.
CM_TO_INCH = 1/2.54


class Browser:
    def __init__(self, msg):
        try:
            self.id = msg['id']
            self.print_url = msg['args'][0]
            self.print_options = msg['args'][1]
            self.load_timeout = msg['args'][2] if len(msg['args']) > 2 else 30
            self.filename = self.print_options.pop('filename', 'print.pdf')
            self.quality = self.print_options.pop('quality', "default")
            if self.quality not in ["original", "screen", "printer", "prepress", "default"]:
                raise Exception(f"Invalid PDF quality '{self.quality}' - must be one of ['original', 'screen', 'printer', 'prepress', 'default']")
        except Exception as e:
            print(f"Invalid request to printing service: {e}")
            self.send_error(e)
        else:
            # TODO: We may want to kill the entire Python process if the timeouts in run_browser aren't enough to guarantee this thread ends.
            threading.Thread(target=self.run_browser).start()

    def run_browser(self):
        port = SystemRandom().randint(10000, 20000)
        print(f"Booting browser on port {port}...")
        # Chrome starts lots of processes. Use os.setsid to start all of them in a new process group, so we can kill them all later.
        with subprocess.Popen(CHROME_SANDBOX_ARGS + ["google-chrome", "--headless", "--disable-gpu", f"--remote-debugging-port={port}"], stderr=subprocess.PIPE, stdout=subprocess.PIPE, preexec_fn=os.setsid) as process:

            try:

                # Wait for Chrome to boot, and devtools to initialise. Only wait ten seconds, otherwise fail.
                for line in non_blocking_readlines(process.stderr, 10):
                    print(f"Chrome: {line}")
                    if line.startswith("DevTools listening"):
                        break
                else:
                    raise Exception("Could not initialise printing service within allowed time.")

                print("Connecting to Chrome...")
                browser = pychrome.Browser(f"http://127.0.0.1:{port}")
                print("Setting up tab...")
                tab = browser.new_tab()
                tab.start()
                tab.Network.enable()
                #tab.Network.clearBrowserCache()
                tab.Runtime.enable()

                ready_to_print = False
                print_error = None

                def handle_console_output(args, **kwargs):
                    nonlocal ready_to_print, print_error
                    print(f"Console: {args}")
                    if args and args[0].get('value', None) == "READY_TO_PRINT": # This string is console.log'd in runner.js.
                        ready_to_print = True
                    elif args and args[0].get('value', None) == "PRINT_ERROR":
                        print_error = args[1].get('value')


                outstanding_network_requests = {}
                def network_request_sent(requestId, **request):
                    if is_interesting_network_request(request):
                        outstanding_network_requests[requestId] = request

                def network_request_failed(requestId, **kwargs):
                    outstanding_network_requests.pop(requestId,None)

                def network_request_finished(requestId, **kwargs):
                    outstanding_network_requests.pop(requestId, None)

                def network_request_cached(requestId, **kwargs):
                    outstanding_network_requests.pop(requestId, None)


                tab.set_listener("Runtime.consoleAPICalled", handle_console_output)
                tab.set_listener("Network.requestWillBeSent", network_request_sent)
                tab.set_listener("Network.loadingFailed", network_request_failed)
                tab.set_listener("Network.loadingFinished", network_request_finished)
                tab.set_listener("Network.requestServedFromCache", network_request_cached) # Not sure if loadingFinished also fires in this case. Do this just to be safe.

                print(f"Navigating to {self.print_url}")
                tab.Page.navigate(url=self.print_url)

                print("Waiting for components to load...")
                loading_start = time.time()
                while time.time() < loading_start + self.load_timeout: # Give things 30 seconds to load.
                    time.sleep(0.01)
                    if ready_to_print and len(outstanding_network_requests) == 0:
                        break
                    if print_error:
                        print(f"It's all gone to pot: {print_error}")
                        self.send_exception(json.loads(print_error))
                        return
                else:
                    if not ready_to_print:
                        raise Exception("Components did not load within allowed time.")
                    else:
                        raise Exception("Network requests did not complete within allowed time.")


                time.sleep(0.5) # Give another half-second after everything seems ready, just to make sure rendering is done.

                print("Printing...")
                pdfdata = tab.Page.printToPDF(displayHeaderFooter=False, printBackground=True, **self.get_pdf_options())
                print("Encoding...")
                pdfbytes = base64.b64decode(pdfdata['data'])
                print(f"Generated PDF of {len(pdfbytes)} bytes")

                browser.close_tab(tab)

                if self.quality != "original":
                    print("Compressing PDF...")                

                    # http://web.mit.edu/ghostscript/www/Ps2pdf.htm
                    
                    # This would be the simple way to do it, but ps2pdf is really slow via stdin/stdout...
                    # ps2pdf = subprocess.run(["ps2pdf", f"-dPDFSETTINGS=/{self.quality}", "-"], input=pdfbytes, capture_output=True)
                    # pdfbytes = ps2pdf.stdout

                    # ... So we use temporary files. Sigh.
                    try:
                        with NamedTemporaryFile(delete=False) as src:
                            src_name = src.name
                            src.write(pdfbytes)
                            src.close()

                            with NamedTemporaryFile(delete=False) as dest:
                                dest_name = dest.name
                                dest.close()

                                ps2pdf = run_ps2pdf(self.quality, src_name, dest_name)

                                if os.path.getsize(dest_name) < len(pdfbytes):
                                    with open(dest_name, "rb") as new_pdf:
                                        pdfbytes = new_pdf.read()

                                    print(f"Compressed PDF to {len(pdfbytes)} bytes")
                                else:
                                    print("No gain from compression")
                    finally:
                        try:
                            os.unlink(src_name)
                            os.unlink(dest_name)
                        except:
                            pass
                    # </silly temp file thing>

                    if ps2pdf.stderr or ps2pdf.returncode != 0:
                        raise Exception(f"Error running ps2pdf: {ps2pdf.stderr.decode('utf-8')}")



                print("Responding...")
                self.send_response(pdfbytes)
                print("PDF returned.")
            except Exception as e:
                import traceback
                traceback.print_exc()
                print(f"It's all gone to pot: {e}")
                self.send_error(e)
            finally:
                for line in non_blocking_readlines(process.stderr, 2):
                    print(f"Chrome: {line}")

                # Don't just kill the subprocess, also terminate all the processes it created.
                os.killpg(os.getpgid(process.pid), 15)
                print("******* Chrome terminated *******")
                cpid = -1
                n_terminated = 0
                while cpid != 0:
                    cpid, _, _ = os.wait4(-1, os.WNOHANG)
                    if cpid != 0:
                        n_terminated += 1
                print(f"{n_terminated} children reaped")

    def get_pdf_options(self):
        # Whitelist these options carefully; they're going into the management end of Chrome
        options = {}
        for k,v in self.print_options.items():
            if k == "landscape":
                options["landscape"] = bool(v)
            elif k == "scale":
                options["scale"] = float(v)
            elif k == "margins":
                if isinstance(v, Number):
                    v = {"top": v, "bottom": v, "left": v, "right": v}

                for side in ["top", "bottom", "left", "right"]:
                    if side in v:
                        options[f"margin{side.capitalize()}"] = float(v[side]) * CM_TO_INCH
            elif k == "page_size":
                if type(v) is str:
                    size = {'A0': (84.1, 118.9),
                            'A1': (59.4, 84.1),
                            'A2': (42.0, 59.4),
                            'A3': (29.7, 42.0),
                            'A4': (21.0, 29.7),
                            'A5': (14.8, 21.0),
                            'A6': (10.5, 14.8),
                            'A7': (7.4, 10.5),
                            'A8': (5.2, 7.4),
                            'A9': (3.7, 5.2),
                            'A10': (2.6, 3.7),
                            'LETTER': (21.6, 27.9)}
                    if v.upper() in size:
                        v = size[v.upper()]
                    else:
                        raise ValueError(f"Unknown page size {v.upper!r} - try specifying (width, height) in cm")

                options["paperWidth"] = float(v[0]) * CM_TO_INCH
                options["paperHeight"] = float(v[1]) * CM_TO_INCH

            else:
                raise ValueError(f"Unknown option {k!r}")

        return options

    # Hacked-up responses
    def send_response(self, data):

        send_to_server({"id": self.id, "response": None,
                        "objects": [{"type": ["DataMedia"], "path": ["response"], "id": "pdf",
                                     "mime-type": "application/pdf", "name": self.filename}]})

        # Copied from _server.serialise()
        l = len(data)
        i = 0
        n = 0
        sent_once = False
        while i < l or not sent_once:
            chunk_len = min(l - i, 65536)

            send_to_server({'type': 'CHUNK_HEADER', 'requestId': self.id, 'mediaId': 'pdf',
                            'chunkIndex': n, 'lastChunk': (i + chunk_len == l)},
                           data[i:(i+chunk_len)])

            i += chunk_len
            n += 1
            sent_once = True

    def send_error(self, e):
        send_to_server({"id": self.id, "error": {"type": type(e).__name__, "message": "PDF generation failed: " + str(e)}})

    def send_exception(self, e):
        send_to_server({"id": self.id, "error": e})


def launch(data):
    print(f"Launching new browser for app {data['app-id']}")
    Browser(data)



import  marshal, os, psutil, random, sys, threading, time
from subprocess import PIPE

# State representing which workers are running here
from anvil_downlink_host import workers_by_id, send_with_header, maybe_quit_if_draining_and_done, \
                                    PopenWithGroupKill, get_demote_fn, TIMEOUT, \
                                    report_worker_stats

CAN_PERSIST = (os.environ.get("DOWNLINK_CAN_PERSIST", "false").lower() in {"true", "1"})

cached_workers = {}
CACHE_LOCK = threading.Lock()

class Worker:
    def __init__(self, first_req_id, enable_profiling=False, app_id=None, app_version=None, cache_key=None, set_timeout=True, task_info=None):
        self.req_ids = {first_req_id}
        self.outbound_ids = {} # Outbound ID -> inbound ID it came from
        self.lock = threading.RLock()
        self.media_tracking = {} # reqID -> (set([mediaId, mediaId, ]), finishedCallback)

        self.proc = PopenWithGroupKill([sys.executable, "-um", "anvil_downlink_worker.full_python_worker"],
                                       bufsize=0, stdin=PIPE, stdout=PIPE, preexec_fn=get_demote_fn(app_id))
        self.proc_info = psutil.Process(self.proc.pid)
        self.task_info = task_info

        workers_by_id[first_req_id] = self
        self.cache_key = cache_key
        if cache_key is not None:
            #print("Saving to cache for %s [version %s]" % (cache_key, app_version))
            with CACHE_LOCK:
                old_version, displaced_worker = cached_workers.get(cache_key, (None, None))
                cached_workers[cache_key] = (app_version, self)
            if displaced_worker is not None:
                print("Displacing persistent worker for %s:\nVersion %s -> %s" % (cache_key, old_version, app_version))
                displaced_worker.drain()
            else:
                print("New persistent worker for %s (version %s)" % (cache_key, app_version))

        self.timed_out = False
        self.timeouts = {}
        self.app_version = app_version
        if cache_key is None and set_timeout:
            self.set_timeout(first_req_id)

        self.start_time = {first_req_id: time.time()}
        self.enable_profiling = {first_req_id: enable_profiling}
        self.killing_task = False

        threading.Thread(target=self.read_loop).start()

    def on_media_complete(self, msg, callback):
        media_ids = set()
        for o in msg.get("objects", []):
            if "DataMedia" in o.get("type", []):
                media_ids.add(o["id"])
        if len(media_ids) == 0:
            callback()
        else:
            #print("Waiting for media for request '%s': %s" % (msg['id'], repr(list(media_ids))))
            self.media_tracking[msg['id']] = (media_ids, callback)

    def transmitted_media(self, request_id, media_id):
        #print("Media complete: '%s', '%s'" % (request_id, media_id))
        if request_id in self.media_tracking:
            media_ids, callback = self.media_tracking[request_id]
            media_ids.discard(media_id)
            if len(media_ids) == 0:
                callback()
                del self.media_tracking[request_id]

    def set_timeout(self, timeout_key):
        #print("Set timeout for %s" % timeout_key)
        if timeout_key in self.timeouts:
            return
        timeout_timer = threading.Timer(TIMEOUT, lambda: self.soft_timeout(timeout_key))
        self.timeouts[timeout_key] = timeout_timer
        timeout_timer.start()

    def clear_timeout(self, timeout_key):
        timeout_timer = self.timeouts.pop(timeout_key, None)
        if timeout_timer is not None:
            timeout_timer.cancel()

    def soft_timeout(self, timeout_key):
        # Something has timed out. If we are cached, drain ourselves nicely before terminating
        print("TIMEOUT for %s" % timeout_key)
        self.timed_out = True
        if self.cache_key is None:
            self.hard_timeout()
        else:
            print("Cached worker %s timed out; draining" % self.cache_key)
            with CACHE_LOCK:
                if self.cache_key in cached_workers and cached_workers[self.cache_key][1] is self:
                    cached_workers.pop(self.cache_key, None)
            self.timeout_timer = threading.Timer(TIMEOUT, self.hard_timeout)
            self.timeout_timer.start()

    def hard_timeout(self):
        print("TIMEOUT TERMINATE FOR %s" % self.req_ids)
        self.timed_out = True
        self.proc.terminate()

    def drain(self):
        # Finish up our execution and time out.
        if self.cache_key is None:
            print("Worker for %s is not cached, ignoring drain request" % self.req_ids)
        elif len(self.req_ids) == 0:
            print("Worker for %s drained, terminating instantly (version %s)" % (self.cache_key, self.app_version))
            try:
                self.proc.terminate()
            except:
                pass
        else:
            print("Worker for %s draining, setting timeout (version %s)" % (self.cache_key, self.app_version))
            self.set_timeout("DRAINED")

    def responded(self, req_id):
        self.clear_timeout(req_id)
        self.req_ids.discard(req_id)
        workers_by_id.pop(req_id, None)
        self.start_time.pop(req_id, None)
        self.enable_profiling.pop(req_id, None)
        if (self.cache_key is None or cached_workers.get(self.cache_key, (None,None))[1] is not self) and len(self.req_ids) == 0:
            # Drain complete; goodbye!
            if self.cache_key is not None:
                print("Cache worker for %s drained (version %s)" % (self.cache_key, self.app_version))
            self.proc.terminate()

        maybe_quit_if_draining_and_done()
        #print("Done @%s -> %s" % (self.cache_key, cached_workers.get(self.cache_key)))

    def kill_background_task(self):
        if self.killing_task:
            return
        self.killing_task = True

        # Request state. If it returns with in 5 seconds, we will die with state, else we hard-kill
        self.timeout_timer = threading.Timer(5, self._hard_kill_background_task)
        self.send({'type': 'GET_TASK_STATE', 'id': 'pre-kill-task-state'})
        self.timeout_timer.start()

    def _hard_kill_background_task(self):
        print("TIMEOUT KILLING BACKGROUND TASK %s" % self.req_ids)
        try:
            self.req_ids.discard('pre-kill-task-state')
            send_with_header({'type': 'NOTIFY_TASK_KILLED', 'id': list(self.req_ids)[0]})
        finally:
            self.hard_timeout()

    # Output gets forwarded straight upstream
    def read_loop(self):
        try:
            while True:
                # marshal.load() takes the GIL, so only do it once we know there's something there to load.
                dummy_char = self.proc.stdout.read(1)
                if len(dummy_char) == 0:
                    break
                msg = marshal.load(self.proc.stdout)
                type = msg.get("type")
                id = msg.get("id") or msg.get("requestId")

                if type == "CALL" or type == "GET_APP":
                    self.outbound_ids[msg["id"]] = msg.get('originating-call')
                    workers_by_id[msg["id"]] = self
                else:
                    if id is None:
                        if "output" in msg:
                            # Output from unknown thread? Broadcast it.
                            print("Broadcasting output from unknown thread: %s" % msg)
                            for i in self.req_ids:
                                msg["id"] = i
                                send_with_header(msg)
                        else:
                            print("Discarding invalid message with no ID: %s" % repr(msg))
                        continue
                    if id not in self.req_ids and id not in self.outbound_ids:
                        print("Discarding invalid message with bogus ID: %s" % repr(msg))
                        if type == "CHUNK_HEADER":
                            print("Discarding binary data chunk")
                            marshal.load(self.proc.stdout)
                        continue

                try:
                    if type == "CHUNK_HEADER":
                        x = marshal.load(self.proc.stdout)
                        send_with_header(msg, x)
                        if msg.get("lastChunk"):
                            self.transmitted_media(msg['requestId'], msg['mediaId'])
                    else:

                        if "response" in msg and self.enable_profiling.get(id):
                            p = msg.get("profile", None)
                            msg["profile"] = {
                                "origin": "Server (Python)",
                                "description": "Downlink dispatch",
                                "start-time": float(self.start_time.get(id, 0)*1000),
                                "end-time": float(time.time()*1000)
                            }
                            if p is not None:
                                msg["profile"]["children"] = [p]

                                for o in msg.get("objects", []):
                                    if o["path"][0] == "profile":
                                        o["path"].insert(1,"children")
                                        o["path"].insert(2, 0)

                        if "response" in msg and msg['id'] == 'pre-kill-task-state':
                            # Special case handling for a "clean" kill (where we manage to recover the state)

                            objects = msg.get('objects', [])
                            for o in objects:
                                if 'path' in o and o['path'][0] == 'response':
                                    o['path'][0] = 'taskState'
                                if 'DataMedia' in o['type']:
                                    msg['objects'] = []
                                    msg['response'] = None
                                    break

                            self.req_ids.discard('pre-kill-task-state')

                            send_with_header({'type': 'NOTIFY_TASK_KILLED', 'id': list(self.req_ids)[0],
                                                         'taskState': msg['response'], 'objects': objects})

                            self.proc.terminate()
                        else:
                            send_with_header(msg)

                    if "response" in msg or "error" in msg:
                        #if statsd and (id in self.start_time):
                        #    statsd.timing('Downlink.WorkerLifetime', (time.time()*1000) - self.start_time.get(id, 0)*1000)
                        self.on_media_complete(msg, lambda: self.responded(id))

                except UnicodeError:
                    send_with_header({"id": id, "error": {"type": "UnicodeError", "message": "This function returned a binary string (not text). If you want to return binary data, use a BlobMedia object instead."}})
                    self.responded(id)

        except EOFError:
            print("EOFError while reading worker stdout. This should not have happened.")
            pass

        finally:
            for i in self.req_ids:
                workers_by_id.pop(i, None)
            for i in self.outbound_ids.keys():
                workers_by_id.pop(i, None)
            rt = self.proc.poll()
            if rt is None:
                self.proc.terminate()
            for _,t in self.timeouts.items():
                t.cancel()
            if self.cache_key is not None and cached_workers.get(self.cache_key, (None,None))[1] is self:
                cached_workers.pop(self.cache_key, None)

            error_id = "".join([random.choice('0123456789abcdef') for x in range(10)])
            for i in self.req_ids:
                if self.timed_out:
                    message = 'Server code took too long'
                    type = "anvil.server.TimeoutError"
                elif rt == -9:
                    message = 'Server code execution process was killed. It may have run out of memory: %s' % (error_id)
                    type = "anvil.server.ExecutionTerminatedError"
                    sys.stderr.write(message + " (IDs %s)\n" % i)
                    sys.stderr.flush()
                else:
                    message = 'Server code exited unexpectedly: %s' % (error_id)
                    type = "anvil.server.ExecutionTerminatedError"
                    sys.stderr.write(message + " (IDs %s)\n" % i)
                    sys.stderr.flush()
                send_with_header({'id': i, 'error':{'type': type, 'message': message}})
            print ("Worker terminated for IDs %s (return code %s)" % (self.req_ids, rt))
            maybe_quit_if_draining_and_done()

    def send(self, msg, bin=False):

        if not bin:
            id = msg.get("id")
            if msg.get("type") in ["CALL", "CALL_WITH_APP", "GET_TASK_STATE", "LAUNCH_REPL", "REPL_COMMAND", "TERMINATE_REPL"]:
                # It's a new request! Start the timeout
                #print ("Setting timeout and routing for new request ID %s" % id)
                workers_by_id[id] = self
                if msg.get("enable-profiling"):
                    self.enable_profiling[id] = True
                    self.start_time[id] = time.time()
                self.req_ids.add(id)
                if msg["type"] != "REPL_COMMAND":
                    self.set_timeout(id)

            elif msg.get("type") == "REPL_KEEPALIVE":
                self.clear_timeout(msg["repl"])
                self.set_timeout(msg["repl"])
                send_with_header({"id": id, "response": None})
                return

            # A horrid hack - a one-char "activation" that's not marshalled, because marshal holds the GIL and Windows doesn't support select() on pipes and urrrggghhh

            d = b"X" + marshal.dumps(msg)
        else:
            d = marshal.dumps(msg)

        self.proc.stdin.write(d)
        self.proc.stdin.flush()

        if not bin:
            def outbound_done():
                self.outbound_ids.pop(id, None)
                workers_by_id.pop(id, None)
                maybe_quit_if_draining_and_done()

            if "response" in msg or "error" in msg:
                self.on_media_complete(msg, outbound_done)

    def get_task_state(self, msg):
        self.send(msg)

    def handle_inbound_message(self, msg, bin=None):
        self.send(msg)
        if bin is not None:
            self.send(bin, bin=True)
            if msg.get("last_chunk"):
                self.transmitted_media(msg.get("requestId"), msg.get("mediaId"))

    report_stats = report_worker_stats


def launch(data):
    type = data.get("type")
    id = data.get("id")

    persist_key = data.get("persist-key")

    start_time = time.time()

    is_background_task = (type in ["LAUNCH_BACKGROUND", "LAUNCH_BACKGROUND_WITH_APP"])
    is_repl_launch = type == "LAUNCH_REPL"
    cache_key = None
    worker = None
    version = None
    supplied_version = data.get("app-version")

    print ("%s '%s' for app '%s' (ID %s)" % ("Launching REPL" if is_repl_launch else "Launching BG task" if is_background_task else "Calling function",
                                             data.get("command", "<no func>"), data.get("app-id", "<unknown>"), id))

    if CAN_PERSIST and not is_background_task and persist_key is not None and "app-id" in data and supplied_version is not None:
        cache_key = repr((data["app-id"], persist_key))
        #print("Attempt persistence: %s" % cache_key)
        version, worker = cached_workers.get(cache_key, (None,None))
        #print("Version %s:\n%s\nvs\n%s" % (("MATCH" if version==supplied_version else "MISMATCH"), version, supplied_version))

    if data.get('command') == "anvil.private.pdf.get_component" and not is_background_task:
        wid = data['args'][0][0]
        worker = workers_by_id.get(wid)
        if worker is None:
            send_with_header({'id': id, 'error': {'message': "No component worker found for print call '%s'" % wid}})
            return

    elif worker is None or version != supplied_version:
        worker = Worker(id, data.get("enable-profiling", False), data.get("app-id", "<unknown>"),
                        cache_key=cache_key, app_version=supplied_version, set_timeout=not is_background_task and not is_repl_launch,
                        task_info={
                            "app_id": data["app-id"],
                            "type": "repl" if is_repl_launch else "background_task" if is_background_task else "persistent_worker" if cache_key else "server_call",
                            "task": data.get("command"),
                            "persist": {"key": persist_key, "version": supplied_version} if cache_key else None,
                        })

    worker.send(data)

    # if statsd:
    #     if is_background_task:
    #         statsd.incr('Downlink.LaunchBackgroundTask')
    #     else:
    #         statsd.incr('Downlink.Call')




import anvil.server

#!defClass(anvil.users,UserExists)!:
class UserExists(anvil.server.AnvilWrappedError):
    pass


#!defClass(anvil.users,AuthenticationFailed)!:
class AuthenticationFailed(anvil.server.AnvilWrappedError):
    pass


#!defClass(anvil.users,EmailNotConfirmed)!:
class EmailNotConfirmed(AuthenticationFailed):
    pass


#!defClass(anvil.users,AccountIsNotEnabled)!:
class AccountIsNotEnabled(AuthenticationFailed):
    pass


#!defClass(anvil.users,TooManyPasswordFailures)!:
class TooManyPasswordFailures(AuthenticationFailed):
    pass


#!defClass(anvil.users,PasswordNotAcceptable)!:
class PasswordNotAcceptable(anvil.server.AnvilWrappedError):
    pass

#!defClass(anvil.users,MFARequired)!:
class MFARequired(AuthenticationFailed):
    pass

#!defClass(anvil.users,PasswordResetRequested)!:
class PasswordResetRequested(anvil.server.AnvilWrappedError):
    pass

